@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import glob
 import os
 import platform
@@ -6,7 +8,6 @@ import subprocess
 import sys
 import sysconfig
 from distutils import log
-from distutils.command.build import build as CommandBuild
 from distutils.errors import (
     CompileError,
     DistutilsExecError,
@@ -14,21 +15,17 @@ from distutils.errors import (
     DistutilsPlatformError,
 )
 from distutils.sysconfig import get_config_var
-from typing import Dict, List, NamedTuple, Optional, cast
+from typing import Dict, List, NamedTuple, Optional, Set, Tuple, cast
 
+from setuptools.command.build import build as CommandBuild  # type: ignore[import]
 from setuptools.command.build_ext import build_ext as CommandBuildExt
 from setuptools.command.build_ext import get_abi3_suffix
 from typing_extensions import Literal
 
+from ._utils import format_called_process_error
 from .command import RustCommand
-from .extension import RustExtension, Strip
-from .utils import (
-    PyLimitedApi,
-    binding_features,
-    get_rust_target_info,
-    get_rust_target_list,
-    split_platform_and_extension,
-)
+from .extension import Binding, RustBin, RustExtension, Strip
+from .rustc_info import get_rust_host, get_rust_target_list, get_rustc_cfgs
 
 
 class build_rust(RustCommand):
@@ -72,7 +69,7 @@ class build_rust(RustCommand):
         super().finalize_options()
 
         if self.plat_name is None:
-            self.plat_name = cast(
+            self.plat_name = cast(  # type: ignore[no-any-unimported]
                 CommandBuild, self.get_finalized_command("build")
             ).plat_name
             assert isinstance(self.plat_name, str)
@@ -130,7 +127,7 @@ class build_rust(RustCommand):
             cross_lib = None
             linker = None
 
-        rustc_cfgs = _get_rustc_cfgs(target_triple)
+        rustc_cfgs = get_rustc_cfgs(target_triple)
 
         env = _prepare_build_environment(cross_lib)
 
@@ -139,17 +136,23 @@ class build_rust(RustCommand):
                 f"can't find Rust extension project file: {ext.path}"
             )
 
+        quiet = self.qbuild or ext.quiet
+        debug = self._is_debug_build(ext)
+
         # Find where to put the temporary build files created by `cargo`
-        target_dir = _base_cargo_target_dir(ext)
+        target_dir = _base_cargo_target_dir(ext, quiet=quiet)
         if target_triple is not None:
             target_dir = os.path.join(target_dir, target_triple)
 
-        quiet = self.qbuild or ext.quiet
-        debug = self._is_debug_build(ext)
         cargo_args = self._cargo_args(
             ext=ext, target_triple=target_triple, release=not debug, quiet=quiet
         )
 
+        rustflags = []
+
+        if linker is not None:
+            rustflags.extend(["-C", "linker=" + linker])
+
         if ext._uses_exec_binding():
             command = [self.cargo, "build", "--manifest-path", ext.path, *cargo_args]
 
@@ -157,14 +160,9 @@ class build_rust(RustCommand):
             rustc_args = [
                 "--crate-type",
                 "cdylib",
+                *ext.rustc_flags,
             ]
 
-            if ext.rustc_flags is not None:
-                rustc_args.extend(ext.rustc_flags)
-
-            if linker is not None:
-                rustc_args.extend(["-C", "linker=" + linker])
-
             # OSX requires special linker arguments
             if sys.platform == "darwin":
                 ext_basename = os.path.basename(self.get_dylib_ext_path(ext, ext.name))
@@ -175,24 +173,12 @@ class build_rust(RustCommand):
                     ]
                 )
 
-            if ext.native:
-                rustc_args.extend(["-C", "target-cpu=native"])
-
             # Tell musl targets not to statically link libc. See
             # https://github.com/rust-lang/rust/issues/59302 for details.
             if rustc_cfgs.get("target_env") == "musl":
                 # This must go in the env otherwise rustc will refuse to build
                 # the cdylib, see https://github.com/rust-lang/cargo/issues/10143
-                MUSL_FLAGS = "-C target-feature=-crt-static"
-                rustflags = env.get("RUSTFLAGS")
-                if rustflags is not None:
-                    env["RUSTFLAGS"] = f"{rustflags} {MUSL_FLAGS}"
-                else:
-                    env["RUSTFLAGS"] = MUSL_FLAGS
-
-                # Include this in the command-line anyway, so that when verbose
-                # logging enabled the user will see that this flag is in use.
-                rustc_args.extend(MUSL_FLAGS.split())
+                rustflags.append("-Ctarget-feature=-crt-static")
 
             command = [
                 self.cargo,
@@ -205,14 +191,26 @@ class build_rust(RustCommand):
                 *rustc_args,
             ]
 
+        if rustflags:
+            existing_rustflags = env.get("RUSTFLAGS")
+            if existing_rustflags is not None:
+                rustflags.append(existing_rustflags)
+            new_rustflags = " ".join(rustflags)
+            env["RUSTFLAGS"] = new_rustflags
+
+            # print RUSTFLAGS being added before the command
+            if not quiet:
+                print(f"[RUSTFLAGS={new_rustflags}]", end=" ", file=sys.stderr)
+
         if not quiet:
             print(" ".join(command), file=sys.stderr)
 
         # Execute cargo
         try:
-            output = subprocess.check_output(command, env=env, encoding="latin-1")
+            stderr = subprocess.PIPE if quiet else None
+            output = subprocess.check_output(command, env=env, stderr=stderr, text=True)
         except subprocess.CalledProcessError as e:
-            raise CompileError(f"cargo failed with code: {e.returncode}\n{e.output}")
+            raise CompileError(format_called_process_error(e))
 
         except OSError:
             raise DistutilsExecError(
@@ -258,14 +256,17 @@ class build_rust(RustCommand):
                         f"unable to find executable '{name}' in '{artifacts_dir}'"
                     )
         else:
-            if sys.platform == "win32" or sys.platform == "cygwin":
+            platform = sysconfig.get_platform()
+            if "win" in platform:
                 dylib_ext = "dll"
-            elif sys.platform == "darwin":
+            elif platform.startswith("macosx"):
                 dylib_ext = "dylib"
+            elif "wasm32" in platform:
+                dylib_ext = "wasm"
             else:
                 dylib_ext = "so"
 
-            wildcard_so = "*{}.{}".format(ext.get_lib_name(), dylib_ext)
+            wildcard_so = "*{}.{}".format(ext.get_lib_name(quiet=quiet), dylib_ext)
 
             try:
                 dylib_paths.append(
@@ -302,7 +303,7 @@ class build_rust(RustCommand):
             if ext._uses_exec_binding():
                 ext_path = build_ext.get_ext_fullpath(module_name)
                 # remove extensions
-                ext_path, _, _ = split_platform_and_extension(ext_path)
+                ext_path, _, _ = _split_platform_and_extension(ext_path)
 
                 # Add expected extension
                 exe = sysconfig.get_config_var("EXE")
@@ -310,7 +311,18 @@ class build_rust(RustCommand):
                     ext_path += exe
 
                 os.makedirs(os.path.dirname(ext_path), exist_ok=True)
-                ext.install_script(module_name.split(".")[-1], ext_path)
+                if isinstance(ext, RustBin):
+                    executable_name = module_name
+                    if exe is not None:
+                        executable_name += exe
+                    wheel = self.get_finalized_command("bdist_wheel")
+                    scripts_dir = os.path.join(
+                        build_ext.build_lib, wheel.data_dir, "scripts"  # type: ignore[attr-defined]
+                    )
+                    os.makedirs(scripts_dir, exist_ok=True)
+                    ext_path = os.path.join(scripts_dir, executable_name)
+                else:
+                    ext.install_script(module_name.split(".")[-1], ext_path)
             else:
                 ext_path = self.get_dylib_ext_path(ext, module_name)
                 os.makedirs(os.path.dirname(ext_path), exist_ok=True)
@@ -374,12 +386,12 @@ class build_rust(RustCommand):
         host_arch = host_platform.rsplit("-", 1)[1]
         # Remove incorrect platform tag if we are cross compiling
         if target_arch and host_arch != target_arch:
-            ext_path, _, extension = split_platform_and_extension(ext_path)
+            ext_path, _, extension = _split_platform_and_extension(ext_path)
             # rust.so, removed platform tag
             ext_path += extension
         return ext_path
 
-    def _py_limited_api(self) -> PyLimitedApi:
+    def _py_limited_api(self) -> _PyLimitedApi:
         bdist_wheel = self.distribution.get_command_obj("bdist_wheel", create=False)
 
         if bdist_wheel is None:
@@ -390,11 +402,12 @@ class build_rust(RustCommand):
 
             bdist_wheel_command = cast(CommandBdistWheel, bdist_wheel)  # type: ignore[no-any-unimported]
             bdist_wheel_command.ensure_finalized()
-            return cast(PyLimitedApi, bdist_wheel_command.py_limited_api)
+            return cast(_PyLimitedApi, bdist_wheel_command.py_limited_api)
 
     def _detect_rust_target(
         self, forced_target_triple: Optional[str] = None
     ) -> Optional["_TargetInfo"]:
+        assert self.plat_name is not None
         cross_compile_info = _detect_unix_cross_compile_info()
         if cross_compile_info is not None:
             cross_target_info = cross_compile_info.to_target_info()
@@ -429,33 +442,23 @@ class build_rust(RustCommand):
                 )
 
         elif forced_target_triple is not None:
-            return _TargetInfo.for_triple(forced_target_triple)
-
-        else:
             # Automatic target detection can be overridden via the CARGO_BUILD_TARGET
             # environment variable or --target command line option
-            return self._detect_local_rust_target()
+            return _TargetInfo.for_triple(forced_target_triple)
 
-    def _detect_local_rust_target(self) -> Optional["_TargetInfo"]:
-        """Attempts to infer the correct Rust target from build environment for
-        some edge cases."""
-        assert self.plat_name is not None
+        # Determine local rust target which needs to be "forced" if necessary
+        local_rust_target = _adjusted_local_rust_target(self.plat_name)
 
-        # If we are on a 64-bit machine, but running a 32-bit Python, then
-        # we'll target a 32-bit Rust build.
-        if self.plat_name == "win32":
-            if _get_rustc_cfgs(None).get("target_env") == "gnu":
-                return _TargetInfo.for_triple("i686-pc-windows-gnu")
-            return _TargetInfo.for_triple("i686-pc-windows-msvc")
-        elif self.plat_name == "win-amd64":
-            if _get_rustc_cfgs(None).get("target_env") == "gnu":
-                return _TargetInfo.for_triple("x86_64-pc-windows-gnu")
-            return _TargetInfo.for_triple("x86_64-pc-windows-msvc")
-        elif self.plat_name.startswith("macosx-") and platform.machine() == "x86_64":
-            # x86_64 or arm64 macOS targeting x86_64
-            return _TargetInfo.for_triple("x86_64-apple-darwin")
-        else:
-            return None
+        # Match cargo's behaviour of not using an explicit target if the
+        # target we're compiling for is the host
+        if (
+            local_rust_target is not None
+            # check for None first to avoid calling to rustc if not needed
+            and local_rust_target != get_rust_host()
+        ):
+            return _TargetInfo.for_triple(local_rust_target)
+
+        return None
 
     def _is_debug_build(self, ext: RustExtension) -> bool:
         if self.release:
@@ -493,7 +496,7 @@ class build_rust(RustCommand):
 
         features = {
             *ext.features,
-            *binding_features(ext, py_limited_api=self._py_limited_api()),
+            *_binding_features(ext, py_limited_api=self._py_limited_api()),
         }
 
         if features:
@@ -502,6 +505,9 @@ class build_rust(RustCommand):
         if ext.args is not None:
             args.extend(ext.args)
 
+        if ext.cargo_manifest_args is not None:
+            args.extend(ext.cargo_manifest_args)
+
         return args
 
 
@@ -509,11 +515,9 @@ def create_universal2_binary(output_path: str, input_paths: List[str]) -> None:
     # Try lipo first
     command = ["lipo", "-create", "-output", output_path, *input_paths]
     try:
-        subprocess.check_output(command)
+        subprocess.check_output(command, text=True)
     except subprocess.CalledProcessError as e:
         output = e.output
-        if isinstance(output, bytes):
-            output = e.output.decode("latin-1").strip()
         raise CompileError("lipo failed with code: %d\n%s" % (e.returncode, output))
     except OSError:
         # lipo not found, try using the fat-macho library
@@ -622,26 +626,11 @@ def _detect_unix_cross_compile_info() -> Optional["_CrossCompileInfo"]:
         linker = None
         linker_args = None
     else:
-        [linker, linker_args] = bldshared.split(maxsplit=1)
+        [linker, _, linker_args] = bldshared.partition(" ")
 
     return _CrossCompileInfo(host_type, cross_lib, linker, linker_args)
 
 
-_RustcCfgs = Dict[str, Optional[str]]
-
-
-def _get_rustc_cfgs(target_triple: Optional[str]) -> _RustcCfgs:
-    cfgs: _RustcCfgs = {}
-    for entry in get_rust_target_info(target_triple):
-        maybe_split = entry.split("=", maxsplit=1)
-        if len(maybe_split) == 2:
-            cfgs[maybe_split[0]] = maybe_split[1].strip('"')
-        else:
-            assert len(maybe_split) == 1
-            cfgs[maybe_split[0]] = None
-    return cfgs
-
-
 def _replace_vendor_with_unknown(target: str) -> Optional[str]:
     """Replaces vendor in the target triple with unknown.
 
@@ -678,16 +667,17 @@ def _prepare_build_environment(cross_lib: Optional[str]) -> Dict[str, str]:
     if cross_lib:
         env.setdefault("PYO3_CROSS_LIB_DIR", cross_lib)
 
+    env.pop("CARGO", None)
     return env
 
 
-def _base_cargo_target_dir(ext: RustExtension) -> str:
+def _base_cargo_target_dir(ext: RustExtension, *, quiet: bool) -> str:
     """Returns the root target directory cargo will use.
 
     If --target is passed to cargo in the command line, the target directory
     will have the target appended as a child.
     """
-    target_directory = ext._metadata()["target_directory"]
+    target_directory = ext._metadata(quiet=quiet)["target_directory"]
     assert isinstance(
         target_directory, str
     ), "expected cargo metadata to contain a string target directory"
@@ -696,7 +686,7 @@ def _base_cargo_target_dir(ext: RustExtension) -> str:
 
 def _is_py_limited_api(
     ext_setting: Literal["auto", True, False],
-    wheel_setting: Optional[PyLimitedApi],
+    wheel_setting: Optional[_PyLimitedApi],
 ) -> bool:
     """Returns whether this extension is being built for the limited api.
 
@@ -719,3 +709,64 @@ def _is_py_limited_api(
 
     # "auto" setting - use whether the bdist_wheel option is truthy.
     return bool(wheel_setting)
+
+
+def _binding_features(
+    ext: RustExtension,
+    py_limited_api: _PyLimitedApi,
+) -> Set[str]:
+    if ext.binding in (Binding.NoBinding, Binding.Exec):
+        return set()
+    elif ext.binding is Binding.PyO3:
+        features = {"pyo3/extension-module"}
+        if ext.py_limited_api == "auto":
+            if isinstance(py_limited_api, str):
+                python_version = py_limited_api[2:]
+                features.add(f"pyo3/abi3-py{python_version}")
+            elif py_limited_api:
+                features.add(f"pyo3/abi3")
+        return features
+    elif ext.binding is Binding.RustCPython:
+        return {"cpython/python3-sys", "cpython/extension-module"}
+    else:
+        raise DistutilsPlatformError(f"unknown Rust binding: '{ext.binding}'")
+
+
+_PyLimitedApi = Literal["cp37", "cp38", "cp39", "cp310", "cp311", "cp312", True, False]
+
+
+def _adjusted_local_rust_target(plat_name: str) -> Optional[str]:
+    """Returns the local rust target for the given `plat_name`, if it is
+    necessary to 'force' a specific target for correctness."""
+
+    # If we are on a 64-bit machine, but running a 32-bit Python, then
+    # we'll target a 32-bit Rust build.
+    if plat_name == "win32":
+        if get_rustc_cfgs(None).get("target_env") == "gnu":
+            return "i686-pc-windows-gnu"
+        else:
+            return "i686-pc-windows-msvc"
+    elif plat_name == "win-amd64":
+        if get_rustc_cfgs(None).get("target_env") == "gnu":
+            return "x86_64-pc-windows-gnu"
+        else:
+            return "x86_64-pc-windows-msvc"
+    elif plat_name.startswith("macosx-") and platform.machine() == "x86_64":
+        # x86_64 or arm64 macOS targeting x86_64
+        return "x86_64-apple-darwin"
+
+    return None
+
+
+def _split_platform_and_extension(ext_path: str) -> Tuple[str, str, str]:
+    """Splits an extension path into a tuple (ext_path, plat_tag, extension).
+
+    >>> _split_platform_and_extension("foo/bar.platform.so")
+    ('foo/bar', '.platform', '.so')
+    """
+
+    # rust.cpython-38-x86_64-linux-gnu.so to (rust.cpython-38-x86_64-linux-gnu, .so)
+    ext_path, extension = os.path.splitext(ext_path)
+    # rust.cpython-38-x86_64-linux-gnu to (rust, .cpython-38-x86_64-linux-gnu)
+    ext_path, platform_tag = os.path.splitext(ext_path)
+    return (ext_path, platform_tag, extension)
